#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "conjgrad.h"

#define min(a,b)  ((a)<(b)?(a):(b))
#define max(a,b)  ((a)<(b)?(b):(a))
#define sign(a,b) ((b)<0?-fabs(a):fabs(a))

typedef struct {
  int dim;
  double *p, *xi;
} cvtstruct;

double *
onedtomany (double x, cvtstruct cvt) 
{
  int i;
  double *p = malloc(cvt.dim * sizeof *p);

  if (!p) {
    fprintf(stderr,"malloc failure\n");
    exit(EXIT_FAILURE);
  }

  for (i = 0; i < cvt.dim; i++) {
    p[i] = cvt.p[i] + cvt.xi[i] * x;
  }

  return p;
}

void 
mnbrak(double *ax, double *bx, double *cx, cvtstruct cvt, 
       double *xtop(double x, cvtstruct cvt),
       double func(int dim, const double *point)) 
{
  const double GOLD = 1.618304, GLIMIT = 100, TINY = 1.e-20;
  double *point;
  double fa, fb, fc;

  fa = func(cvt.dim, point = xtop(*ax, cvt)); free(point);
  fb = func(cvt.dim, point = xtop(*bx, cvt)); free(point);

  if (fb > fa) {
    double dum = *ax; *ax = *bx; *bx = dum;
    dum = fb; fb = fa; fa = dum;
  }

  *cx = *bx + GOLD * (*bx - *ax);
  fc = func(cvt.dim, point = xtop(*cx, cvt)); free(point);

  while (fb >= fc) {
    double r = (*bx - *ax) * (fb - fc);
    double q = (*bx - *cx) * (fb - fa);
    double u = *bx - ( (*bx - *cx)*q - (*bx-*ax)*r ) /
                    (2 * sign(max(fabs(q-r),TINY),q-r));
    double fu;
    double ulim = *bx + GLIMIT * (*cx - *bx);
    if ( (*bx - u)*(u - *cx) > 0 ) {
      fu = func(cvt.dim, point = xtop(u, cvt)); free(point);
      if (fu < fc) {
	*ax = *bx; fa = fb;
	*bx = u; fb = fu;
	return;
      } else if (fu > fb) {
	*cx = u; fc = fu;
	return;
      }
      u = *cx + GOLD * (*cx - *bx);
      fu = func(cvt.dim, point = xtop(u, cvt)); free(point);
    } else if ( (*cx-u)*(u-ulim) > 0 ) {
      fu = func(cvt.dim, point = xtop(u, cvt)); free(point);
      if (fu < fc) {
	*bx = *cx; *cx = u; u = *cx + GOLD * (*cx - *bx);
	fb = fc; fc = fu;
	fu = func(cvt.dim, point = xtop(u, cvt)); free(point);
      }
    } else if ( (u-ulim)*(ulim-*cx) >= 0 ) {
      u = ulim;
      fu = func(cvt.dim, point = xtop(u, cvt)); free(point);
    } else {
      u = *cx + GOLD*(*cx-*bx);
      fu = func(cvt.dim, point = xtop(u, cvt)); free(point);
    }

    *ax = *bx; *bx = *cx; *cx = u;
    fa = fb; fb = fc; fc = fu;
  }

  return;
}

funcmin 
brent(double ax, double bx, double cx, cvtstruct cvt,
      double *xtop(double  x, cvtstruct cvt),
      double     f(int dim, const double *p),
      double tol) 
{
  funcmin retval;
  double a = min(ax,bx);
  double b = max(ax,bx);
  double v = bx, w = v, x = v, e = 0;
  double *point;
  double fx = f(cvt.dim, point = xtop(x, cvt));
  double fv = fx, fw = fx;
  double d = e;

  const int ITMAX = 100;
  const double CGOLD = 0.3819660, ZEPS = 1.e-10;

  free(point);
  retval.dim  = cvt.dim;

  for (retval.iter = 0; retval.iter < ITMAX; retval.iter++) {
    double xm = (a+b)/2;
    double tol1 = tol*fabs(x)+ZEPS;
    double tol2 = 2 * tol1;
    int dogolden = 1;
    double u, fu;

    if (fabs(x-xm) <= (tol2 - (b-a)/2)) break;

    if (fabs(e) > tol1) {
      double r = (x-w)*(fx-fv);
      double q = (x-v)*(fx-fw);
      double p = (x-v)*q - (x-w)*r;
      double etemp;

      q = 2 * (q-r);
      if (q > 0) p = -p; else q = -q;
      
      etemp = e;
      e = d;
      if (fabs(p) < fabs(q*etemp/2) &&
	  p < q*(b-x)) {
	d = p/q;
	u = x+d;
	if (u-a < tol2 ||
	    b-u < tol2 ) d = sign(tol1,xm-x);
	dogolden = 0;
      }
    }

    if (dogolden) d = CGOLD * (e = (x >= xm ? a : b) - x);

    u  = x + ((fabs(d) >= tol1)?d:sign(tol1,d));
    fu = f(cvt.dim, point = xtop(u, cvt)); free(point);
    if (fu <= fx) {
      if (u>=x) a = x; else b = x;
      v = w; fv = fw;
      w = x; fw = fx;
      x = u; fx = fu;
    } else {
      if (u < x) a = u; else b = u;
      if ( fu <= fw || w == x ) {
	v = w; fv = fw;
	w = u; fw = fu;
      } else if ( fu <= fv || v == x || v == w ) {
	v = u; fv = fu;
      }
    }
  }
  if ( retval.iter > ITMAX ) {
    fprintf(stderr, "Too many iterations\n");
    exit(EXIT_FAILURE);
  }

  retval.point = xtop(x,cvt);
  retval.fret = fx;

  return retval;
}


funcmin 
frprmn(const int dim, const double ftol, const double *initpoint,
       double   func(int dim, const double point[]),
       double *dfunc(int dim, const double point[])) 
{
  const int ITMAX = 200; const double EPS = 1.e-10;

  double *p = malloc(dim * sizeof *p);
  double *g = malloc(dim * sizeof *g);
  double *h = malloc(dim * sizeof *h);

  double fp, *xi;

  int j, its;
  funcmin retval;

  /* Check allocation */
  if (!p || !g || !h) {
    fprintf(stderr, "Out of memory\n");
    exit(EXIT_FAILURE);
  }

  /* Initialize */
  retval.dim   = dim;
  retval.point = p;
  memcpy(p,initpoint,dim * sizeof *p);
  fp =  func(dim, p); xi = dfunc(dim, p);
  retval.fret = fp;
  for (j = 0; j < dim; j++) {
    h[j] = g[j] = xi[j] = -xi[j];
  }

  /* Iterate over line minimizations */
  for (its = 0; its < ITMAX; its++) {
    double ax = 0, bx = 1, cx;
    double gg=0, dgg=0, gam;
    cvtstruct cvt;
    const double brenttol = 1.e-4;
    double *point;
    int ok = 0;

    cvt.dim = dim; cvt.p   = p; cvt.xi  = xi;

    while (!ok) {
      int ii; double sum = 0;
      point = onedtomany(bx, cvt);
      ok = 1;
      for (ii = 0; ii<dim; ii++) {
	ok &= point[ii] > 0;
	if (ii < 3) sum += point[ii];
      }
      ok &= sum < 1;
      if (!ok) {
	bx /= 2;
	if (bx < 1.e-20) 
	  if (bx > 0) bx = -1.e-5;
	  else {
	    retval.point = onedtomany(ax,cvt); retval.iter = 0; free(cvt.p); 
	    free(g); free(h); 
	    printf("ERROR: RETURNING BECAUSE CAN'T MOVE AT ALL!\n");
	    return retval;
	  }
	printf("Readjusting bx to %g\n", bx/10);
      }
    }
    bx /= 10;

    mnbrak(&ax, &bx, &cx, cvt, onedtomany, func);
    p = (retval = brent(ax, bx, cx, cvt, onedtomany, func, brenttol)).point;
    retval.iter = its; free(cvt.p);

    if ( 2*fabs(retval.fret-fp) <= ftol*(fabs(retval.fret)+fabs(fp)+EPS) ) {
      free(g); free(h); return retval;
    }

    fp = func(dim, p); free(xi); xi = dfunc(dim, p);
    for (j=0; j<dim; j++) {
      gg  += g[j]*g[j];
      dgg += (xi[j]+g[j])*xi[j];
    }
    if (gg == 0) {
      free(g); free(h); return retval;
    }

    gam = dgg/gg;
    for (j=0; j<dim; j++) {
      xi[j] = h[j] = (g[j] = -xi[j]) + gam*h[j];
    }
  }

  fprintf(stderr,"Too many iterations\n");
  exit(EXIT_FAILURE);

}

#ifdef CHECK

// ++pac #include "conjgrad.h"

double chisqr(int dim, const double point[]) {
  double result=0;
  int i;
  for (i = 0; i<dim; i++) {
    result += ((point[i] - i) * (point[i] - i) + 3) *
              ((point[i] - i) * (point[i] - i) + 3) ;
  }
  return result;
}

double *dchisqr(int dim, const double point[]) {
  double *result = malloc(dim * sizeof *result);
  int i;
  
  if (!result) {
    fprintf(stderr, "malloc failure\n");
    exit (EXIT_FAILURE);
  }
  
  for (i = 0; i < dim; i++) {
    result[i] = 2 * ((point[i] - i) * (point[i] - i) + 3) *
                    2 * (point[i] - i);
  }

  return result;
}

int main(void) {

  double initpoint[5] = {5, 2, 32, 23, 12};

  funcmin result = frprmn(5, 1.e-10, initpoint, chisqr, dchisqr);

  printf("Dimension of problem = %d\n",  result.dim);
  printf("Number of iterations = %d\n",  result.iter);
  printf("Final point = %f %f %f %f %f\n", result.point[0], result.point[1], result.point[2], 
	                                   result.point[3], result.point[4]);
  printf("Function at final point = %f\n", result.fret);

  return 0;

}

#endif

